# Alternatives:
##  - watchfiles.watch(): needs `pip3 install watchfiles`, which installs [anyio, idna, sniffio].  Seems fine.
##  - watchexec: needs `cargo` to install

import time, os, sys, subprocess as subp
from typing import List,Dict,Iterator

try: import watchfiles
except ImportError: have_watchfiles = False
else: have_watchfiles = True


def run(args:List[str]) -> None:
    if not args:
        print('Usage:')
        print('  kpa watch ./serve.py --port=8000   # Re-runs when serve.py changes')
        print('  kpa watch utils.py -- kpa lint utils.py  # Re-runs when utils.py changes')
        sys.exit(1)
    elif '--' in args:
        dashdash_index = args.index('--')
        filepaths = args[:dashdash_index]
        cmd = args[dashdash_index+1:]
    else:
        filepaths = [args[0]]
        cmd = args
        if os.path.exists(filepaths[0]) and not filepaths[0].startswith(('/','./')):
            cmd[0] = './'+cmd[0]

    print(f'Watching {repr(filepaths)} and running {cmd}\n')
    for _ in yield_when_files_update(filepaths, and_also_immediately=True):
        print('======>', cmd)
        subp.run(cmd)
        print('.\n')



def yield_when_files_update(filepaths:List[str], and_also_immediately:bool = False) -> Iterator[None]:
    assert isinstance(filepaths, list)
    if and_also_immediately: yield
    if have_watchfiles:
        for changeset in watchfiles.watch(*filepaths, raise_interrupt=False):
            yield
    else:
        mtimes = get_mtimes(filepaths)
        while True:
            time.sleep(0.5)
            new_mtimes = get_mtimes(filepaths)
            if new_mtimes != mtimes:
                print("Files were updated:", [p for p in filepaths if mtimes[p]!=new_mtimes[p]])
                yield
            mtimes = new_mtimes

def get_mtimes(filepaths:List[str]) -> Dict[str,float]: return {p:get_mtime(p) for p in filepaths}
def get_mtime(filepath:str) -> float: return os.stat(filepath).st_mtime



#!/usr/bin/env python3

import sys
from pathlib import Path
from typing import List


def run(argv:List[str]) -> None:
    command = argv[0] if argv else ''
    if command == 'tests/lint.sh': create_lint_sh()
    elif command.endswith('.sh'): create_bash(command)
    else:
        print(f"Unknown command: {repr(command)}")
        print('Usage:')
        print('  kpa skel tests/lint.sh')
        print('  kpa skel $script.sh')


lint_sh_text = '''\
#!/bin/bash
## This script does simple static-checking of this codebase.
## Run `./tests/lint.sh install` to make this run every time you run `git commit`.
set -euo pipefail
readlinkf() { perl -MCwd -le 'print Cwd::abs_path shift' "$1"; }
SCRIPTDIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
exists() { type -t "$1" >/dev/null; }
print_and_run() { echo "=> $@"; "$@"; echo; }
cd "$SCRIPTDIR/.."

if [[ ${1:-} == install ]]; then
    print_and_run python3 -m pip install flake8 mypy kpa
    echo -e "#!/bin/bash\n./tests/lint.sh" > .git/hooks/pre-commit
    print_and_run chmod a+x .git/hooks/pre-commit
    echo "=> Installed pre-commit hook!"
    exit 0
fi

if [[ ${1:-} == watch ]]; then
    kpa lint --watch
else
    kpa lint
fi

echo FINISHED
'''

def create_lint_sh() -> None:
    lint_sh_path = Path('tests/lint.sh')
    lint_sh_path.parent.mkdir(exist_ok=True)
    if not lint_sh_path.exists(): lint_sh_path.write_text(lint_sh_text)

bash_text = '''\
#!/bin/bash
set -euo pipefail
readlinkf() { perl -MCwd -le 'print Cwd::abs_path shift' "$1"; }
SCRIPTDIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
exists() { type -t "$1" >/dev/null; }
print_and_run() { echo "=> $@"; "$@"; echo; }
cd "$SCRIPTDIR"

'''

def create_bash(filepath:str) -> None:
    p = Path(filepath)
    p.parent.mkdir(exist_ok=True)
    if p.exists() and '-f' not in sys.argv: print(f"Cannot create {filepath} because it already exists!  Consider using `-f`."); exit(1)
    p.write_text(bash_text)



#!/usr/bin/env python3

from pathlib import Path
import pathlib, sys, argparse, os
from typing import Optional


def main() -> None:
    command = sys.argv[1] if sys.argv[1:] else ''

    if command in ['lint', 'l']:
        from .dev_utils import lint_cli
        lint_cli(sys.argv[2:])

    elif command in ['lw', 'lint-watch', 'wl', 'watch-lint']:
        from .dev_utils import lint_cli
        lint_cli(sys.argv[2:] + ['--watch'])

    elif command in ['watch', 'w']:
        from .watcher import run as watcher_run
        watcher_run(sys.argv[2:])

    elif command in ["pip-find-updates", 'pfu']:
        from .pip_utils import run as pfu_run
        pfu_run(sys.argv[2:])

    elif command == 'skel':
        from .skel import run as skel_run
        skel_run(sys.argv[2:])

    elif command == 'serve-status-code':
        from .http_server import serve, status_code_server
        serve(status_code_server)

    elif command == 'serve-dir':
        from .http_server import serve, magic_directory_server
        serve(magic_directory_server)

    elif command == 'serve-redirect':
        from .http_server import serve, make_redirect_server
        port = int(sys.argv[2])
        target_base_url = sys.argv[3]
        serve(make_redirect_server(target_base_url), port=port)

    elif command in ['term-color', 'termcolor']:
        from .terminal_utils import termcolor
        def r(num): return '#' if num is None else str(num%10)
        print('    # ' + ' '.join('{bg:2}'.format(bg=bg) for bg in range(50)))
        for fg in [None]+list(range(0,25)):
            print('{fg:<2} '.format(fg=(fg if fg is not None else '#')) +
                  ' '.join(termcolor(r(fg)+r(bg), fg, bg) for bg in [None]+list(range(50))))

    else:
        if sys.argv[1:]: print('unknown command:', sys.argv[1:])
        from .version import version
        print(f'kpa version {version}\n')
        print('available commands:\n  kpa termcolor\n  kpa status-code-server\n  kpa redirect-server\n  kpa pip-find-updates')



import time, random, sys, argparse, os, subprocess as subp
from pathlib import Path
from functools import wraps
from typing import Optional,Iterator,List

class ExecutableNotFound(Exception): pass


def lint_cli(argv:List[str]) -> None:
    parser = argparse.ArgumentParser(prog='kpa lint')
    parser.add_argument('files', nargs='*', help='If no files are passed, this uses **/*.py')
    #parser.add_argument('--no-mypy-cache', action='store_true', help="Don't make .mypy_cache/")  # Conflicts with `--install-types`.  Consider using `--cache-dir=/tmp/{slugify(abspaths(args.files))}`.
    parser.add_argument('--run-rarely', action='store_true', help="Only when file is modified in last 30 seconds, or otherwise 1%% of the time")
    parser.add_argument('--extra-flake8-ignores', help="Extra errors/warnings for flake8 to ignore")
    parser.add_argument('--venv-bin-dir', help="A path to venv/bin that has flake8 or mypy")
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--watch', action='store_true', help='Watch files, and re-lint when files are updated')
    args = parser.parse_args(argv)
    if not args.files: args.files = list(get_all_py_files())
    _lint_cli(args)
def _lint_cli(args:argparse.Namespace) -> None:
    if args.watch:
        from .watcher import yield_when_files_update
        args.watch = False
        for _ in yield_when_files_update(args.files, and_also_immediately=True):
            if len(args.files)<5: print(f'=====> linting {args.files}...')
            else: print(f'=====> linting {len(args.files)} files...')
            _lint_cli(args)
            print('.\n')
        exit(0)

    if args.run_rarely:
        seconds_since_last_change = time.time() - max(Path(path).stat().st_mtime for path in args.files)
        if seconds_since_last_change > 30 and random.random() > 0.01: exit(0)

    def find_exe(name:str) -> str:
        for path in find_exe_options(name):
            if os.path.exists(path): return path
        print(f"[Failed to find {name}]")
        raise ExecutableNotFound()
    def find_exe_options(name:str) -> Iterator[str]:
        if args.venv_bin_dir: yield f'{args.venv_bin_dir}/{name}'
        yield f'{os.path.dirname(sys.executable)}/{name}'
        yield f'{os.path.dirname(sys.argv[0])}/{name}'
        try: yield subp.check_output(['which',name], stderr=subp.DEVNULL).decode().strip()
        except Exception: pass
        yield f'venv/bin/{name}'
        for file in args.files: yield f'{os.path.dirname(os.path.abspath(file))}/venv/bin/{name}'
    def print_and_run(cmd:List[str]) -> bool:  # Returns whether it succeeded
        if args.verbose: print('=>', cmd)
        p = subp.run(cmd)
        if p.returncode != 0: print(f"\n{cmd[0]} failed"); return False
        return True

    flake8_ignore = 'B007,E116,E124,E126,E127,E128,E129,E201,E202,E203,E221,E222,E225,E226,E227,E228,E231,E241,E251,E261,E265,E266,E301,E302,E303,E305,E306,E401,E402,E501,E701,E702,E704,F401,F811,W292,W293,W391,W504'
    if args.extra_flake8_ignores: flake8_ignore += ',' + args.extra_flake8_ignores
    try: flake8_exe = find_exe('flake8')
    except ExecutableNotFound: print("flake8 not found")
    else:
        if not print_and_run([flake8_exe, '--show-source', f'--ignore={flake8_ignore}', *args.files]): return

    try: mypy_exe = find_exe('mypy')
    except ExecutableNotFound: print("mypy not found")
    else: print_and_run([mypy_exe, '--pretty', '--ignore-missing-imports', '--non-interactive', '--install-types', *args.files])


def lint(filepath:str = '', make_cache:bool = True, run_rarely:bool = False) -> None:
    if run_rarely:
        seconds_since_last_change = time.time() - Path(filepath).stat().st_mtime
        if seconds_since_last_change > 30 and random.random() > 0.01:
            return  # Don't run
    lint_flake8(filepath)
    lint_mypy(filepath, make_cache=make_cache)
run = lint

def lint_flake8(filepath:str = '') -> None:
    try: flake8_exe = find_exe('flake8', filepath=filepath)
    except ExecutableNotFound: return None
    p = subp.run([flake8_exe, '--show-source', '--ignore=E501,E302,E251,E701,E226,E305,E225,E261,E231,E301,E306,E402,E704,E265,E201,E202,E303,E124,E241,E127,E266,E221,E126,E129,F811,E222,E401,E702,E203,E116,E228,W504,W293,B007,W391,F401,W292,E227,E128', filepath])
    if p.returncode != 0: sys.exit(1)

def lint_mypy(filepath:str = '', make_cache:bool = True) -> None:
    try: mypy_exe = find_exe('mypy', filepath=filepath)
    except ExecutableNotFound: return None
    cmd = [mypy_exe, '--pretty', '--ignore-missing-imports']
    if not make_cache: cmd.append('--cache-dir=/dev/null')
    if filepath: cmd.append(filepath)
    p = subp.run(cmd)
    if p.returncode != 0: sys.exit(1)

def find_exe(name:str, filepath:str = '') -> str:
    for path in find_exe_options(name, filepath=filepath):
        if os.path.exists(path): return path
    print(f"[Failed to find {name}]")
    raise ExecutableNotFound()
def find_exe_options(name:str, filepath:str = '') -> Iterator[str]:
    try: yield subp.check_output(['which',name], stderr=subp.DEVNULL).decode().strip()
    except Exception: pass
    yield f'venv/bin/{name}'
    if filepath: yield f'{os.path.dirname(os.path.abspath(filepath))}/venv/bin/{name}'

def get_all_py_files(directory_:Optional[str] = None) -> Iterator[str]:
    directory = Path(directory_) if directory_ else Path().absolute()
    for filepath in directory.rglob('*.py'):
        if '#' in filepath.name: continue
        rel_path = filepath.relative_to(directory)
        if any(str(name).startswith('.') and str(name)!='.' for name in rel_path.parents): continue
        if 'build/lib/' in str(rel_path): continue
        yield str(filepath)


def get_size(obj, seen:Optional[set] = None) -> int:
    """Recursively calculates bytes of RAM taken by object"""
    # From https://code.activestate.com/recipes/577504/ and https://github.com/bosswissam/pysize/blob/master/pysize.py
    if seen is None: seen = set()
    size = sys.getsizeof(obj)
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)  # Mark as seen *before* recursing to handle self-referential objects

    if isinstance(obj, dict):
        size += sum(get_size(v, seen) for v in obj.values())
        size += sum(get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(get_size(i, seen) for i in obj)

    if hasattr(obj, '__slots__'):  # obj can have both __slots__ and __dict__
        size += sum(get_size(getattr(obj, s), seen) for s in obj.__slots__ if hasattr(obj, s))

    return size
